;; This work is licensed under the Creative Commons
;; Attribution-NonCommercial-ShareAlike License. To view a copy of this
;; license, visit http://creativecommons.org/licenses/by-nc-sa/2.0/ or
;; send a letter to Creative Commons, 559 Nathan Abbott Way, Stanford,
;; California 94305, USA.
;;
;; The original author of this code is Arthur Nunes-Harwitt
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;; Lexical Analysis
;;;;;;;;;;;;;;;;;;;;;;;;;;


(define symbol<-string string->symbol)
(define string<-symbol symbol->string)

(define (lowercase<-string str)
  (list->string (map char-downcase (string->list str))))

(define (logo-list? x) (or (null? x) (pair? x)))

(define (make-tword sym str) (vector 'tword sym str))

(define (tword? x) (and (vector? x) (eq? (vector-ref x 0) 'tword)))
(define (symbol<-tword w) (vector-ref w 1))
(define (string<-tword w) (vector-ref w 2))

(define (make-name str)
  (make-tword (symbol<-string (lowercase<-string str)) str))
(define (logo-name? x) 
  (and (tword? x)
       (not (var-tword? x))
       (not (literal-tword? x))
       (not (memq (symbol<-tword x) *operator-list*))
       (not (oparen-tword? x))
       (not (cparen-tword? x))))

(define symbol<-logo-name symbol<-tword)
(define string<-logo-name string<-tword)

(define (logo-name<-symbol sym)
  (make-tword sym (string<-symbol sym)))

(define *operator-list* '(+ - * / ^ < > =))

(define (make-arithmetic-op sym) (make-tword sym (string<-symbol sym)))
(define (logo-arithmetic-op? x)
  (and (tword? x) 
       (not (var-tword? x))
       (not (literal-tword? x))
       (memq (symbol<-tword x) *operator-list*)))
(define symbol<-logo-arithmetic-op symbol<-tword)

(define (make-var str)
  (make-tword (symbol<-string (lowercase<-string str))
              (string-append ":" str)))
(define (logo-var? x)
  (and (tword? x) 
       (< 0 (string-length (string<-tword x)))
       (char=? (string-ref (string<-tword x) 0) #\:)))
(define (var-tword? x)
  (and (< 0 (string-length (string<-tword x)))
       (char=? (string-ref (string<-tword x) 0) #\:)))
(define symbol<-logo-var symbol<-tword)
(define string<-logo-var string<-tword)
(define (logo-var<-symbol sym)
  (make-tword sym (string-append ":" (string<-symbol sym))))

(define (make-literal-word str)
  (make-tword (symbol<-string (lowercase<-string str)) 
              (string-append "\"" str)))
(define (logo-literal-word? x)
  (and (tword? x) 
       (< 0 (string-length (string<-tword x)))
       (char=? (string-ref (string<-tword x) 0) #\")))
(define (literal-tword? x)
  (and (< 0 (string-length (string<-tword x)))
       (char=? (string-ref (string<-tword x) 0) #\")))
(define symbol<-logo-literal-word symbol<-tword)
(define string<-logo-literal-word string<-tword)


(define *open-paren-token* (symbol<-string "("))
(define *close-paren-token* (symbol<-string ")"))

(define (make-open-paren) (make-tword *open-paren-token* "("))
(define (logo-open-paren? x) 
  (and (tword? x) (eq? (symbol<-tword x) *open-paren-token*)))
(define (oparen-tword? x) (eq? (symbol<-tword x) *open-paren-token*))

(define (make-close-paren) (make-tword *close-paren-token* ")"))
(define (logo-close-paren? x)
  (and (tword? x) (eq? (symbol<-tword x) *close-paren-token*)))
(define (cparen-tword? x) (eq? (symbol<-tword x) *close-paren-token*))

(define (make-number str)
  (let ((start (string-ref str 0)))
    (if (or (char=? start #\+)
            (char=? start #\-))
      (string->number str)
      (string->number (string-append "+" str)))))


(define (logo-lex str)
  (let ((index 0))
    (define (peek-char) (string-ref str index))
    (define (empty?) (>= index (string-length str)))
    (define (advance-char!) (set! index (+ index 1)))
    (define (match-char? c) (char=? (peek-char) c))
    (define (end-of-token?) (or (empty?)
                                (char-whitespace? (peek-char))
                                (match-char? #\()
                                (match-char? #\))
                                (match-char? #\[)
                                (match-char? #\])))

    (define (gen-lex-sublist)
      (advance-char!)
      (let ((result (gen-lex-list)))
        (if (or (empty?) (not (match-char? #\])))
          (logo-error "Unexpected end of list")
          (begin
            (advance-char!)
            result) )))

    (define (gen-lex-var)
      (advance-char!)
      (let ((start index))
        (let loop ((end start))
          (if (end-of-token?)
            (make-var (substring str start end))
            (begin
              (advance-char!)
              (loop (+ end 1)))))))

    (define (gen-lex-name)
      (let ((start index))
        (let loop ((end start))
          (if (end-of-token?)
            (make-name (substring str start end))
            (begin
              (advance-char!)
              (loop (+ end 1)))))))


    (define (gen-try-number)
      (let ((start index))
        
        (define (state0)
          (cond ((or (match-char? #\+) (match-char? #\-))
                 (advance-char!) (state1))
                ((or (char-numeric? (peek-char)) 
                     (match-char? #\.))
                 (advance-char!) (state2))
                (else (set! index start) (gen-lex-name))))
        
        (define (state1)
          (cond ((end-of-token?) (set! index start) (gen-lex-plus-or-minus))
                ((match-char? #\.) (advance-char!) (state5))
                ((char-numeric? (peek-char)) (advance-char!) (state2))
                (else (set! index start) (gen-lex-name))))
        
        (define (state2)
          (cond ((end-of-token?) (make-number (substring str start index)))
                ((char-numeric? (peek-char)) (advance-char!) (state2))
                ((match-char? #\.) (advance-char!) (state6))
                ((match-char? #\/) (advance-char!) (state3))
                ((or (match-char? #\+) (match-char? #\-))
                 (advance-char!) (state10))
                ((or (match-char? #\e) (match-char? #\E))
                 (advance-char!) (state7))
                ((or (match-char? #\i) (match-char? #\I)) 
                 (advance-char!) (state12))
                (else (set! index start) (gen-lex-name))))
        
        (define (state3)
          (cond ((end-of-token?) (set! index start) (gen-lex-name))
                ((char-numeric? (peek-char)) (advance-char!) (state4))
                (else (set! index start) (gen-lex-name))))
    
        (define (state4)
          (cond ((end-of-token?) (make-number (substring str start index)))
                ((char-numeric? (peek-char)) (advance-char!) (state4))
                ((match-char? #\+) (advance-char!) (state10))
                ((or (match-char? #\i) (match-char? #\I)) 
                 (advance-char!) (state12))
                (else (set! index start) (gen-lex-name))))
        
        (define (state5)
          (cond ((end-of-token?) (set! index start) (gen-lex-name))
                ((char-numeric? (peek-char)) (advance-char!) (state6))
                (else (set! index start) (gen-lex-name))))
        
        (define (state6)
          (cond ((end-of-token?) (make-number (substring str start index)))
                ((char-numeric? (peek-char)) (advance-char!) (state6))
                ((or (match-char? #\e) (match-char? #\E))
                 (advance-char!) (state7))
                ((match-char? #\+) (advance-char!) (state10))
                ((or (match-char? #\i) (match-char? #\I)) 
                 (advance-char!) (state12))
                (else (set! index start) (gen-lex-name))))

        (define (state7)
          (cond ((end-of-token?) (set! index start) (gen-lex-name))
                ((char-numeric? (peek-char)) (advance-char!) (state9))
                ((or (match-char? #\+) (match-char? #\-)) 
                 (advance-char!) (state8))
                (else (set! index start) (gen-lex-name))))
        
        (define (state8)
          (cond ((end-of-token?) (set! index start) (gen-lex-name))
                ((char-numeric? (peek-char)) (advance-char!) (state9))
                (else (set! index start) (gen-lex-name))))

        (define (state9) (state4))
        
        (define (state10)
          (cond ((end-of-token?) (set! index start) (gen-lex-name))
                ((match-char? #\.) (advance-char!) (state16))
                ((char-numeric? (peek-char)) (advance-char!) (state11))
                (else (set! index start) (gen-lex-name))))
        
        (define (state11)
          (cond ((end-of-token?) (set! index start) (gen-lex-name))
                ((char-numeric? (peek-char)) (advance-char!) (state11))
                ((match-char? #\.) (advance-char!) (state17))
                ((match-char? #\/) (advance-char!) (state13))
                ((or (match-char? #\i) (match-char? #\I)) 
                 (advance-char!) (state12))
                ((or (match-char? #\e) (match-char? #\E))
                 (advance-char!) (state18))
                (else (set! index start) (gen-lex-name))))

        (define (state12)
          (cond ((end-of-token?) (make-number (substring str start index)))
                (else (set! index start) (gen-lex-name))))
        
        (define (state13)
          (cond ((end-of-token?) (set! index start) (gen-lex-name))
                ((char-numeric? (peek-char)) (advance-char!) (state14))
                (else (set! index start) (gen-lex-name))))
        
        (define (state14)
          (cond ((end-of-token?) (make-number (substring str start index)))
                ((char-numeric? (peek-char)) (advance-char!) (state14))
                ((or (match-char? #\i) (match-char? #\I)) 
                 (advance-char!) (state15))
                (else (set! index start) (gen-lex-name))))


        (define (state15) (state12))

        (define (state16)
          (cond ((end-of-token?) (set! index start) (gen-lex-name))
                ((char-numeric? (peek-char)) (advance-char!) (state17))
                (else (set! index start) (gen-lex-name))))
        
        (define (state17)
          (cond ((end-of-token?) (make-number (substring str start index)))
                ((char-numeric? (peek-char)) (advance-char!) (state17))
                ((or (match-char? #\e) (match-char? #\E))
                 (advance-char!) (state18))
                ((or (match-char? #\i) (match-char? #\I)) 
                 (advance-char!) (state15))
                (else (set! index start) (gen-lex-name))))

        (define (state18)
          (cond ((end-of-token?) (set! index start) (gen-lex-name))
                ((char-numeric? (peek-char)) (advance-char!) (state20))
                ((or (match-char? #\+) (match-char? #\-)) 
                 (advance-char!) (state19))
                (else (set! index start) (gen-lex-name))))

        (define (state19)
          (cond ((end-of-token?) (set! index start) (gen-lex-name))
                ((char-numeric? (peek-char)) (advance-char!) (state20))
                (else (set! index start) (gen-lex-name))))
        
        (define (state20) (state14))

        (state0)))

    (define (gen-lex-plus-or-minus)
      (cond ((match-char? #\+)
             (advance-char!)
             (make-arithmetic-op '+))
            ((match-char? #\-)
             (advance-char!)
             (make-arithmetic-op '-))
            (else (error "This shouldn't happen - lexeme" (peek-char)))))
        
        

    (define (gen-lex-times)
      (let ((start index))
        (advance-char!)
        (if (end-of-token?)
          (make-arithmetic-op '*)
          (begin
            (set! index start)
            (gen-lex-name)))))

    (define (gen-lex-divide)
      (let ((start index))
        (advance-char!)
        (if (end-of-token?)
          (make-arithmetic-op '/)
          (begin
            (set! index start)
            (gen-lex-name)))))

    (define (gen-lex-power)
      (let ((start index))
        (advance-char!)
        (if (end-of-token?)
          (make-arithmetic-op '^)
          (begin
            (set! index start)
            (gen-lex-name)))))

    (define (gen-lex-less)
      (let ((start index))
        (advance-char!)
        (if (end-of-token?)
          (make-arithmetic-op '<)
          (begin
            (set! index start)
            (gen-lex-name)))))


    (define (gen-lex-greater)
      (let ((start index))
        (advance-char!)
        (if (end-of-token?)
          (make-arithmetic-op '>)
          (begin
            (set! index start)
            (gen-lex-name)))))


    (define (gen-lex-equal)
      (let ((start index))
        (advance-char!)
        (if (end-of-token?)
          (make-arithmetic-op '=)
          (begin
            (set! index start)
            (gen-lex-name)))))


    (define (gen-lex-literal-word)
      (advance-char!)
      (let ((start index))
        (let loop ((end start))
          (if (end-of-token?)
            (make-literal-word (substring str start end))
            (begin
              (advance-char!)
              (loop (+ end 1)))))))


    (define (gen-lex-list)
      (cond ((empty?) '())
            ((char-whitespace? (peek-char))
             (advance-char!)
             (gen-lex-list))
            ((match-char? #\[) 
             (cons (gen-lex-sublist) (gen-lex-list)))
            ((match-char? #\]) '())
            ((match-char? #\:) 
             (cons (gen-lex-var) (gen-lex-list)))
            ((match-char? #\*) 
             (cons (gen-lex-times) (gen-lex-list)))
            ((match-char? #\/) 
             (cons (gen-lex-divide) (gen-lex-list)))
            ((match-char? #\^) 
             (cons (gen-lex-power) (gen-lex-list)))
            ((match-char? #\<) 
             (cons (gen-lex-less) (gen-lex-list)))
            ((match-char? #\>) 
             (cons (gen-lex-greater) (gen-lex-list)))
            ((match-char? #\=) 
             (cons (gen-lex-equal) (gen-lex-list)))
            ((or (char-numeric? (peek-char))
                 (match-char? #\+)
                 (match-char? #\-)
                 (match-char? #\.))
             (cons (gen-try-number) (gen-lex-list)))
            ((match-char? #\")
             (cons (gen-lex-literal-word) (gen-lex-list)))
            ((match-char? #\()
             (advance-char!)
             (cons (make-open-paren) (gen-lex-list)))
            ((match-char? #\)) 
             (advance-char!)
             (cons (make-close-paren) (gen-lex-list)))
            (else (cons (gen-lex-name) (gen-lex-list)))))
            
    (gen-lex-list)))

