;; This work is licensed under the Creative Commons
;; Attribution-NonCommercial-ShareAlike License. To view a copy of this
;; license, visit http://creativecommons.org/licenses/by-nc-sa/2.0/ or
;; send a letter to Creative Commons, 559 Nathan Abbott Way, Stanford,
;; California 94305, USA.
;;
;; The original author of this code is Arthur Nunes-Harwitt
;;
;;; Grammar for Logo
;;;
;;   TS -> empty | TOP TS
;;   TOP -> "to" name var* token* "end" | R
;;   R -> E = E | E < E | E > E | E
;;   E -> T E'
;;   E' -> + T E' | - T E' | empty
;;   T -> F T'
;;   T' -> * F T' | / F T' | empty
;;   F -> - F | P ^ F | P
;;   P -> number | list | lit-word 
;;      | var | name_n R1 ... Rn | ( R ) | (name R*)
;; output and stop should be part of the syntax
;; it would also be useful to have if, repeat, and other lazy forms
;; as part of the syntax
;; procedure body cannot be parsed until runtime since
;; the arity may not be known.  Parsing actually occurs a 
;; little too early, since procedures can create procedures.




(define *arity-table* '())

(define (get-logo-arity name)
  (let* ((sym (symbol<-logo-name name))
	 (result (assq sym *arity-table*)))
    (if result (cdr result) #f)))


(define (put-logo-arity name n)
  (let ((sym (symbol<-logo-name name)))
    (set! *arity-table* (cons (cons sym n) *arity-table*))))


(define (arg1 x) (vector-ref x 1))
(define (arg2 x) (vector-ref x 2))
(define (arg3 x) (vector-ref x 3))

(define (make-sum t1 t2) (vector 'sum t1 t2))
(define (sum? x)
  (and (vector? x) (eq? (vector-ref x 0) 'sum)))

(define (make-diff t1 t2) (vector 'diff t1 t2))
(define (diff? x)
  (and (vector? x) (eq? (vector-ref x 0) 'diff)))

(define (make-prod f1 f2) (vector 'prod f1 f2))
(define (prod? x)
  (and (vector? x) (eq? (vector-ref x 0) 'prod)))

(define (make-quo f1 f2) (vector 'quo f1 f2))
(define (quo? x)
  (and (vector? x) (eq? (vector-ref x 0) 'quo)))

(define (make-neg f) (vector 'neg f))
(define (neg? x)
  (and (vector? x) (eq? (vector-ref x 0) 'neg)))

(define (make-power p f) (vector 'power p f))
(define (power? x)
  (and (vector? x) (eq? (vector-ref x 0) 'power)))

(define (make-equal e1 e2) (vector 'equal e1 e2))
(define (syntax-equal? x)
  (and (vector? x) (eq? (vector-ref x 0) 'equal)))

(define (make-less e1 e2) (vector 'less e1 e2))
(define (syntax-less? x)
  (and (vector? x) (eq? (vector-ref x 0) 'less)))

(define (make-greater e1 e2) (vector 'greater e1 e2))
(define (syntax-greater? x)
  (and (vector? x) (eq? (vector-ref x 0) 'greater)))

(define (make-empty-seq) (vector 'empty-seq))
(define (empty-seq? x)
  (and (vector? x) (eq? (vector-ref x 0) 'empty-seq)))

(define (make-seq e1 e2) (vector 'seq e1 e2))
(define (seq? x)
  (and (vector? x) (eq? (vector-ref x 0) 'seq)))

(define (make-call name args) (vector 'call name args))
(define (call? x)
  (and (vector? x) (eq? (vector-ref x 0) 'call)))

(define (make-stop) (vector 'stop))
(define (stop? x)
  (and (vector? x) (eq? (vector-ref x 0) 'stop)))

(define (make-output e) (vector 'output e))
(define (output? x)
  (and (vector? x) (eq? (vector-ref x 0) 'output)))

(define (make-proc name formals body) (vector 'proc name formals body))
(define (proc? x)
  (and (vector? x) (eq? (vector-ref x 0) 'proc)))

(define (make-catch e1 e2) (vector 'catch e1 e2))
(define (catch? x)
  (and (vector? x) (eq? (vector-ref x 0) 'catch)))

(define (make-throw e) (vector 'throw e))
(define (throw? x)
  (and (vector? x) (eq? (vector-ref x 0) 'throw)))

(define (make-if e1 e2) (vector 'if e1 e2))
(define (if? x)
  (and (vector? x) (eq? (vector-ref x 0) 'if)))

(define (make-ifelse e1 e2 e3) (vector 'ifelse e1 e2 e3))
(define (ifelse? x)
  (and (vector? x) (eq? (vector-ref x 0) 'ifelse)))

(define (make-repeat e1 e2) (vector 'repeat e1 e2))
(define (repeat? x)
  (and (vector? x) (eq? (vector-ref x 0) 'repeat)))

(define (make-make e1 e2) (vector 'make e1 e2))
(define (make? x)
  (and (vector? x) (eq? (vector-ref x 0) 'make)))




(define (parse logo-lex)
  
  (define (peek-token) 
    (if (empty?)
      (logo-error "Unexpected end of input")
      (car logo-lex)))
  (define (empty?) (null? logo-lex))
  (define (advance!) (set! logo-lex (cdr logo-lex)))

  (define (get-ts)
    (if (empty?)
      (make-empty-seq)
      (let ((e (get-top)))
        (make-seq e (get-ts)))))

  (define (get-top)
    (let ((look-ahead (peek-token)))
      (if (and (logo-name? look-ahead)
               (eq? 'to (symbol<-logo-name look-ahead)))
        (begin (advance!) (get-proc))
        (get-rel))))

  (define (get-proc)
    (let ((proc-name (peek-token)))
      (if (logo-name? proc-name)
        (begin
          (advance!)
          (let loop ((vars '())
                     (cur-var (peek-token)))
            (if (logo-var? cur-var)
              (begin
                (advance!)
                (loop (cons cur-var vars) (peek-token)))
              (let ((formals (reverse vars)))
                (let get-body ((body '())
                               (look-ahead (peek-token)))
                  (cond ((empty?) (logo-error "Expected 'end'"))
                        ((and (logo-name? look-ahead)
                              (eq? 'end (symbol<-logo-name look-ahead)))
                         (advance!) 
                         (put-logo-arity proc-name (length formals))
                         (make-proc proc-name formals 
                                              (reverse body)))
                        (else 
                         (advance!)
                         (get-body (cons look-ahead body)
                                        (peek-token)))) )))))
        (logo-error "Expected a procedure name, not" proc-name))))
                                               

  (define (get-seq)
    (if (empty?)
      (make-empty-seq)
      (let ((r (get-rel)))
        (make-seq r (get-seq)))))

  (define (get-rel)
    (let ((e1 (get-exp)))
      (if (empty?)
        e1
        (let ((look-ahead (peek-token)))
          (if (logo-arithmetic-op? look-ahead)
            (cond ((eq? '= (symbol<-logo-arithmetic-op look-ahead))
                   (advance!)
                   (let ((e2 (get-exp)))
                     (make-equal e1 e2)))
                  ((eq? '< (symbol<-logo-arithmetic-op look-ahead))
                   (advance!)
                   (let ((e2 (get-exp)))
                     (make-less e1 e2)))
                  ((eq? '> (symbol<-logo-arithmetic-op look-ahead))
                   (advance!)
                   (let ((e2 (get-exp)))
                     (make-greater e1 e2)))
                  (else  e1))
            e1)))))
      

  (define (get-exp)
    (get-exp-prime (get-term)))

  (define (get-exp-prime term1)
    (if (empty?) 
      term1
      (let ((look-ahead (peek-token)))
        (if (logo-arithmetic-op? look-ahead)
          (cond ((eq? '+ (symbol<-logo-arithmetic-op look-ahead))
                 (advance!)
                 (let ((term2 (get-term)))
                   (get-exp-prime (make-sum term1 term2))))
                ((eq? '- (symbol<-logo-arithmetic-op look-ahead))
                 (advance!)
                 (let ((term2 (get-term)))
                   (get-exp-prime (make-diff term1 term2))))
                (else term1))
          term1))))

  (define (get-term)
    (get-term-prime (get-factor)))

  (define (get-term-prime factor1)
    (if (empty?) 
      factor1
      (let ((look-ahead (peek-token)))
        (if (logo-arithmetic-op? look-ahead)
          (cond ((eq? '* (symbol<-logo-arithmetic-op look-ahead))
                 (advance!)
                 (let ((factor2 (get-factor)))
                   (get-term-prime (make-prod factor1 factor2))))
                ((eq? '/ (symbol<-logo-arithmetic-op look-ahead))
                 (advance!)
                 (let ((factor2 (get-factor)))
                   (get-term-prime (make-quo factor1 factor2))))
                (else factor1))
          factor1))))

  (define (get-factor)
    (let ((look-ahead (peek-token)))
      (if (and (logo-arithmetic-op? look-ahead)
               (eq? '- (symbol<-logo-arithmetic-op look-ahead)))
        (begin
          (advance!)
          (make-neg (get-factor)))
        (let ((prim (get-prim)))
          (if (empty?)
            prim
            (let ((look-ahead (peek-token)))
              (if (and (logo-arithmetic-op? look-ahead)
                       (eq? '^ (symbol<-logo-arithmetic-op look-ahead)))
                (begin
                  (advance!)
                  (make-power prim (get-factor)))
                prim)))))))

  (define (get-prim)
    (let ((look-ahead (peek-token)))
      (cond ((number? look-ahead)
             (advance!)
             look-ahead)
            ((logo-list? look-ahead)
             (advance!)
             look-ahead)
            ((logo-literal-word? look-ahead)
             (advance!)
             look-ahead)
            ((logo-var? look-ahead)
             (advance!)
             look-ahead)
            ((logo-name? look-ahead)
             (advance!)
             (get-bounded-call look-ahead))
            ((logo-open-paren? look-ahead)
             (advance!)
             (get-paren-exp))
            (else (logo-error "Unexpected token" look-ahead)))))

  (define (get-bounded-call name)
    (let ((sym (symbol<-logo-name name)))
      (cond ((eq? sym 'stop)
             (make-stop))
            ((or (eq? sym 'op)
                 (eq? sym 'output))
             (make-output (get-rel)))
            ((eq? sym 'catch)
             (let* ((arg1 (get-rel))
                    (arg2 (get-rel)))
               (make-catch arg1 arg2)))
            ((eq? sym 'throw)
             (make-throw (get-rel)))
            ((eq? sym 'if)
             (let* ((arg1 (get-rel))
                    (arg2 (get-rel)))
               (make-if arg1 arg2)))
            ((eq? sym 'ifelse)
             (let* ((arg1 (get-rel))
                    (arg2 (get-rel))
                    (arg3 (get-rel)))
               (make-ifelse arg1 arg2 arg3)))
            ((eq? sym 'repeat)
             (let* ((arg1 (get-rel))
                    (arg2 (get-rel)))
               (make-repeat arg1 arg2)))
            ((eq? sym 'make)
             (let* ((arg1 (get-rel))
                    (arg2 (get-rel)))
               (make-make arg1 arg2)))
            (else
             (let ((arity (get-logo-arity name)))
               (if (integer? arity)
                 (let loop ((arity arity)
                            (arg-list '()))
                   (if (= arity 0)
                     (make-call name (reverse arg-list))
                     (loop (- arity 1) (cons (get-rel) arg-list))))
                 (logo-error "I don't know how to" name)))) )))


  (define (get-paren-exp)
    (let ((look-ahead (peek-token)))
      (if (logo-name? look-ahead)
        (begin (advance!) (get-unbounded-call look-ahead))
        (let ((e (get-rel)))
          (if (logo-close-paren? (peek-token))
            (begin (advance!) e)
            (logo-error "Close parenthesis expected"))))))

  (define (get-unbounded-call name)
    (let ((sym (symbol<-logo-name name)))
      (cond ((eq? sym 'stop)
             (if (logo-close-paren? (peek-token))
               (begin (advance!) (make-stop))
               (logo-error "Close parenthesis expected")))
            ((or (eq? sym 'op)
                 (eq? sym 'output))
             (let ((arg (get-rel)))
               (if (logo-close-paren? (peek-token))
                 (begin (advance!) (make-output arg))
                 (logo-error "Close parenthesis expected"))))
            ((eq? sym 'catch)
             (let* ((arg1 (get-rel))
                    (arg2 (get-rel)))
               (if (logo-close-paren? (peek-token))
                 (begin (advance!) (make-catch arg1 arg2))
                 (logo-error "Close parenthesis expected"))))
            ((eq? sym 'throw)
             (let ((arg (get-rel)))
               (if (logo-close-paren? (peek-token))
                 (begin (advance!) (make-throw arg))
                 (logo-error "Close parenthesis expected"))))
            ((eq? sym 'if)
             (let* ((arg1 (get-rel))
                    (arg2 (get-rel)))
               (if (logo-close-paren? (peek-token))
                 (begin (advance!) (make-if arg1 arg2))
                 (logo-error "Close parenthesis expected"))))
            ((eq? sym 'ifelse)
             (let* ((arg1 (get-rel))
                    (arg2 (get-rel))
                    (arg3 (get-rel)))
               (if (logo-close-paren? (peek-token))
                 (begin (advance!) (make-ifelse arg1 arg2 arg3))
                 (logo-error "Close parenthesis expected"))))
            ((eq? sym 'repeat)
             (let* ((arg1 (get-rel))
                    (arg2 (get-rel)))
               (if (logo-close-paren? (peek-token))
                 (begin (advance!) (make-repeat arg1 arg2))
                 (logo-error "Close parenthesis expected"))))
            ((eq? sym 'make)
             (let* ((arg1 (get-rel))
                    (arg2 (get-rel)))
               (if (logo-close-paren? (peek-token))
                 (begin (advance!) (make-make arg1 arg2))
                 (logo-error "Close parenthesis expected"))))
            (else
             (let loop ((arg-list '()))
               (if (logo-close-paren? (peek-token))
                 (begin (advance!) (make-call name (reverse arg-list)))
                 (loop (cons (get-rel) arg-list))))))))

  (get-ts))

